;;; -*- mode: TI-Asm; ti-asm-listing-file: "units.lst" -*-
;;;
;;;	Unit Operations Flash App for TI-83+
;;; 
;;;	Copyright 2003 Benjamin Moody
;;;
;;;  This program is free software; you can redistribute
;;;  and/or modify it under the terms of the GNU General
;;;  Public License as published by the Free Software
;;;  Foundation; either version 2 of the License, or (at
;;;  your option) any later version.
;;;
;;;  This program is distributed in the hope that it will be
;;;  useful, but WITHOUT ANY WARRANTY; without even the
;;;  implied warranty of MERCHANTABILITY or FITNESS FOR A
;;;  PARTICULAR PURPOSE. See the GNU General Public License
;;;  for more details.
;;;
;;;  You should have received a copy of the GNU General
;;;  Public License along with this program; if not, write
;;;  to the Free Software Foundation, Inc., 59 Temple Place
;;;  - Suite 330, Boston, MA 02111 USA.
;;;

;;; display.asm: code related to Disp_, Text(, and Output(


;;; Arguments aren't passed normally.  There are a few ways I've thought of
;;; dealing with this.  I've chosen the method here because it's simple to
;;; implement, and will only cause problems if you're being an idiot and using
;;; seriously bad syntax.

;;; Thanks go to Martin for the pointers, though I decided not to use the
;;; _IncFetch method after all.

;;; So what do we do? When a Disp token is found, it's replaced with an
;;; overloaded sin(, DS style.

;;; This code is heavily based on Omnicalc's; thanks go to Brandon and Michael
;;; for this one!

ParserHook_Preparse:
	;; Only need to change stuff if we're parsing a program
	ld a,(OP1)
	cp EquObj
;	jp z,SecondPreparser
	ret z
	cp NewEquObj
;	jp z,SecondPreparser
	ret z
	B_CALL _ChkFindSym
	jp c,RetZ		; Whoa. It seems that 0 isn't always preparse
	ld a,b			; after all.
	or a
	ret nz
	ex de,hl
	push hl
	 ld c,(hl)
	 inc hl
	 ld b,(hl)
	 inc hl
	 xor a
	 ld (preparseChangedP),a
PreparserLoop:
	 ld a,b
	 or c
	 jp z,PreparserDone1
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tString
	 jr z,PreparserQuote
	 cp tDisp
	 jp z,PreparserDisp
	 cp tText
	 jp z,PreparserText
	 cp tOutput
	 jr z,PreparserOutput
	 cp $bb
	 jr z,PreparserDetached
	 B_CALL _IsA2ByteTok
	 jr nz,PreparserLoop
	 inc hl
	 dec bc
	 jr PreparserLoop
PreparserQuote:
	 ld a,b
	 or c
	 jp z,PreparserDone1
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tString
	 jr z,PreparserLoop
	 cp tEnter
	 jr z,PreparserLoop
	 cp tStore
	 jr z,PreparserLoop
	 B_CALL _IsA2ByteTok
	 jr nz,PreparserQuote
	 inc hl
	 dec bc
	 jr PreparserQuote
PreparserDetached:
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tLastDetached+1
	 jr nc,PreparserLoop
	 sub tFirstDetached
	 jr c,PreparserLoop
	 pop de
	push bc
	 push hl
	  push hl
	   push de
	    dec hl
	    ld (hl),tReal
	    ld hl,3
	    push af
	     B_CALL _EnoughMem
	     jp c,ErrMemory
	     pop af
	    pop hl
	   ld c,(hl)
	   inc hl
	   ld b,(hl)
	   inc bc
	   inc bc
	   inc bc
	   ld (hl),b
	   dec hl
	   ld (hl),c
	  ex (sp),hl
	   ex de,hl
	   push af
	    B_CALL _InsertMem
	    pop af
	   ex de,hl
	   ld b,'0'-1
PreparserDiv10:
	   inc b
	   sub 10
	   jr nc,PreparserDiv10
	   add a,10+'0'
	   ld (hl),b
	   inc hl
	   ld (hl),a
	   inc hl
	   ld (hl),tComma
	   pop de
	  pop hl
	 pop bc
	push de
	 inc hl
	 inc hl
	 inc hl
	 ld a,1
	 ld (preparseChangedP),a
	 jp PreparserLoop

PreparserOutput:
	 ld a,t2
	 jr PreparserSin
PreparserDisp:
	 ld a,t0
	 jr PreparserSin
PreparserText:
	 ld a,t1
PreparserSin:
	 pop de
	push bc
	 push hl
	  push hl
	   push de
	    dec hl
	    ld (hl),tSin
	    ld hl,2
	    push af
	     B_CALL _EnoughMem
	     jp c,ErrMemory
	     pop af
	    pop hl
	   ld c,(hl)
	   inc hl
	   ld b,(hl)
	   inc bc
	   inc bc
	   ld (hl),b
	   dec hl
	   ld (hl),c
	  ex (sp),hl
	   ex de,hl
	   push af
	    B_CALL _InsertMem
	    pop af
	   ex de,hl
	   ld (hl),a
	   inc hl
	   ld (hl),tComma
	   pop de
	  pop hl
	 pop bc
	push de
	 inc hl
	 inc hl
	 ld a,1
	 ld (preparseChangedP),a
	 jp PreparserLoop
PreparserDone1:
	 pop de
	ld a,(preparseChangedP)
	or a
;	jr z,PreparserNoChange
	ret z

	rst rPushRealO1
;	call SecondPreparser
;	jr nz,PreparserSkipParse
	ld hl,ParserHook_2
	in a,(6)
	B_CALL _EnableParserHook
	ld hl,Preparser_ErrorHandler
	call APP_PUSH_ERRORH
	B_CALL _ParseInp
	call APP_POP_ERRORH
PreparserSkipParse:
	call OP1ToOP6
	call Postparser
	call OP6ToOP1
	or 1
	ret

; PreparserNoChange:
; 	call SecondPreparser
; 	ret

Postparser:
	call PopOP1
	B_CALL _ChkFindSym
	ex de,hl
	push hl
	 ld c,(hl)
	 inc hl
	 ld b,(hl)
	 inc hl
PostparserLoop:
	 ld a,b
	 or c
	 jp z,PostparserDone
	 bit 7,b		; did we scan past the end?
	 jp nz,PostparserDone
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tString
	 jr z,PostparserQuote
	 cp tSin
	 jr z,PostparserSin
	 cp $bb
	 jr z,PostparserDetached
	 B_CALL _IsA2ByteTok
	 jr nz,PostparserLoop
	 inc hl
	 dec bc
	 jr PostparserLoop
PostparserQuote:
	 ld a,b
	 or c
	 jp z,PostparserDone
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tString
	 jr z,PostparserLoop
	 cp tEnter
	 jr z,PostparserLoop
	 cp tStore
	 jr z,PostparserLoop
	 B_CALL _IsA2ByteTok
	 jr nz,PostparserQuote
	 inc hl
	 dec bc
	 jr PostparserQuote
PostparserDetached:
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tReal
	 jr nz,PostparserLoop
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp t9+1
	 jr nc,PostparserLoop
	 sub t0
	 jr c,PostparserLoop
	 ld d,a
	 add a,a
	 add a,a
	 add a,d
	 add a,a
	 ld d,a
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp t9+1
	 jr nc,PostparserLoop
	 sub t0
	 jr c,PostparserLoop
	 add a,d
	 ld d,a
	 ld a,(hl)
	 inc hl
	 dec bc
	 cp tComma
	 jr nz,PostparserLoop
	 bit 7,b
	 jr nz,PostparserDone
	 dec hl
	 dec hl
	 dec hl
	 ld a,d
	 add a,tFirstDetached
	 pop de
	push bc
	 push hl
	  push hl
	   push de
	    dec hl
	    ld (hl),a
	    pop hl
	   ld c,(hl)
	   inc hl
	   ld b,(hl)
	   dec bc
	   dec bc
	   dec bc
	   ld (hl),b
	   dec hl
	   ld (hl),c
	   ld de,3
	  ex (sp),hl
	   B_CALL _DelMem
	   pop de
	  pop hl
	 pop bc
	push de
	 jp PostparserLoop
PostparserSin:
	 ld a,(hl)
	 cp t0
	 ld d,tDisp
	 jr z,PostparserSinOK
	 cp t1
	 ld d,tText
	 jr z,PostparserSinOK
	 cp t2
	 jp nz,PostparserLoop
	 ld d,tOutput
PostparserSinOK:
	 inc hl
	 ld a,(hl)
	 cp tComma
	 jp nz,PostparserLoop
	 ld a,d
	 dec hl
	 pop de
	push bc
	 push hl
	  push hl
	   push de
	    dec hl
	    ld (hl),a
	    pop hl
	   ld c,(hl)
	   inc hl
	   ld b,(hl)
	   dec bc
	   dec bc
	   ld (hl),b
	   dec hl
	   ld (hl),c
	   ld de,2
	  ex (sp),hl
	   B_CALL _DelMem
	   pop de
	  pop hl
	 pop bc
	push de
	 dec bc
	 dec bc
	 jp PostparserLoop
PostparserDone:
	 pop de
	ret

Preparser_ErrorHandler:
	push af
	call Postparser
	pop af
	B_CALL _JError

; SecondPreparser:
; 	call LoadAppVar_RKH
; 	jp c,RetZ
; 	xor a
; 	push af
; 	push hl
; 	push hl
; 	push hl
; 	jp SecondParserHook

ParserHook_2:
	.db $83
	push hl
	ld hl,ParserHook
	ld (parserHookPtr),hl
	pop hl
	or a
	ret z
	jp ParserHook_Normal

Sin:	pop hl
	ld a,h
	or a
	jr nz,Sin_Multi
	ld a,l
	dec a
	ret z
Sin_Multi:
	push hl
	B_CALL _PushOP1
	pop hl
	push hl
	;; First argument is the special function to be run
	call ParserGetFirst
	dec hl
	ld a,(OP1+2)
	or a
	jr z,Disp
	cp $10
	jp z,Text
	cp $20
	jr z,Output
	pop hl
	or 1
	ret

;;; We've managed to trick TIOS into reading the arguments to Disp and putting
;;; them on the FPS for us. Now we just have to display them.
Disp:
DispLoop:
	ld a,h
	or l
	jr z,DispDone
	call ParserGetFirst
	dec hl
	push hl
	call SetOP1Units_NO_STRING
	jr nz,Disp_NotUnit
	call ConvertToString
	call FillString
Disp_NotUnit:
	B_CALL _FormDisp	; Very convenient! This just Disp_'s OP1
	pop hl
	jr DispLoop

Output:	ld a,h
	or a
	jp nz,ErrArgument
	ld a,l
	cp 3			; Only 3 arguments to Output(, thankfully
	jp nz,ErrArgument
	call ParserGetFirst
	dec hl
	push hl
	 B_CALL _ErrD_OP1NotPosInt
	 B_CALL _ConvOP1
	 ld a,d
	 or a
	 jp nz,ErrDomain
	 ld a,(winTop)
	 add a,e
	 ld e,a
	 ld a,(winBtm)
	 cp e
	 jp c,ErrDomain
	 pop hl
	push de
	 call ParserGetFirst
	 dec hl
	 push hl
	  B_CALL _ErrD_OP1NotPosInt
	  B_CALL _ConvOP1
	  ld a,d
	  or a
	  jp nz,ErrDomain
	  ld a,e
	  cp 17
	  jp nc,ErrDomain
	  pop hl
	 pop de
	ld d,a
	push de
	 call ParserGetFirst
	 call SetOP1Units_NO_STRING
	 call z,ConvertToString
	 pop hl
	B_CALL _OutputExpr
	jr DispDone

TextDone:
	B_CALL _GrBufCpy
	res textWrite,(iy+sGrFlags)
	res fracDrawLFont,(iy+fontFlags)
DispDone:
	pop hl
	call ParserGetFirst
	ld (FPS),de
	res numOP1,(iy+ParsFlag2)
	or 1
	ret

Text:	ld a,(flags+fmtFlags)
	ld (flags+fmtOverride),a

	set textWrite,(iy+sGrFlags)
	push hl
	B_CALL _PDspGrph
	pop hl

	call ParserGetFirst
	dec hl
	ld a,(OP1)
	and $1f
	jp nz,PopRetZ
	;; Mysterious undocumented feature: if the first argument to Text( is
	;; negative, it will display the given text in large font. In this
	;; case the second and third arguments specify the pixel coordinates.
	res fracDrawLFont,(iy+fontFlags)
	ld a,(OP1)
	and $80
	jr z,Text_NotBig
	set fracDrawLFont,(iy+fontFlags)
	call ParserGetFirst
	dec hl
Text_NotBig:
	ld a,h
	or a
	jr nz,Text_ManyArgs
	ld a,l
	cp 2
	jp c,PopRetZ
Text_ManyArgs:

	ld a,(OP1)
	and $1f
	or a
	jp nz,PopRetZ
	push hl
	B_CALL _ConvOP1
	pop hl
	ld a,d
	or a
	jp nz,PopRetZ
	ld a,e
	cp 58
	jp nc,PopRetZ
	ld (penRow),a

	call ParserGetFirst
	dec hl
	ld a,(OP1)
	and $1f
	or a
	jp nz,PopRetZ
	push hl
	B_CALL _ConvOP1
	pop hl
	ld a,d
	or a
	jp nz,PopRetZ	
	ld a,e
	cp 95
	jp nc,PopRetZ
	ld (penCol),a

	push hl
Text_Loop:
	pop hl
	ld a,h
	or l
	jp z,TextDone
	call ParserGetFirst
	dec hl
	push hl
	;; Valid types for display: string, real, complex
	ld a,(OP1)
	or a
	jr z,Text_Real
	cp CplxObj
	jr z,Text_Cplx
	cp StrngObj
	jr z,Text_String
	;; TIOS doesn't handle others; it might be nice for us to do so,
	;; but that would take some work...
	B_CALL _ErrDataType
Text_Real:
	ld a,16
	B_CALL _FormBase
	ld hl,OP3
	push hl
	add hl,bc
	ld (hl),0
	pop hl
	B_CALL _VPutS
	jr Text_Loop
Text_Cplx:
	;; This is actually the only case we have to worry about.
	call SetOP1Units
 	jr nz,Text_Cplx_ReallyComplex
	call ConvertToString
Text_String:
	B_CALL _ChkFindSym
	jp c,ErrUndefined
	ld a,b
	or a
	jp nz,ErrArchived
	ex de,hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
Text_String_Loop:
	ld a,b
	or c
	jr z,Text_Loop
	push hl
	push bc
	B_CALL _Get_Tok_Strng
	ld hl,OP3
	ld b,a
	B_CALL _VPutSN
	pop bc
	pop hl
	ld a,(hl)
	B_CALL _IsA2ByteTok
	jr nz,Text_String_1B
	inc hl
	dec bc
Text_String_1B:
	inc hl
	dec bc
	jr Text_String_Loop
Text_Cplx_ReallyComplex:
	B_CALL _FormDCplx
	ld hl,OP3
	B_CALL _VPutS
	jr Text_Loop
