;;; -*- mode: TI-Asm; ti-asm-listing-file: "units.lst" -*-
;;;
;;;	Unit Operations Flash App for TI-83+
;;; 
;;;	Copyright 2003 Benjamin Moody
;;;
;;;  This program is free software; you can redistribute
;;;  and/or modify it under the terms of the GNU General
;;;  Public License as published by the Free Software
;;;  Foundation; either version 2 of the License, or (at
;;;  your option) any later version.
;;;
;;;  This program is distributed in the hope that it will be
;;;  useful, but WITHOUT ANY WARRANTY; without even the
;;;  implied warranty of MERCHANTABILITY or FITNESS FOR A
;;;  PARTICULAR PURPOSE. See the GNU General Public License
;;;  for more details.
;;;
;;;  You should have received a copy of the GNU General
;;;  Public License along with this program; if not, write
;;;  to the Free Software Foundation, Inc., 59 Temple Place
;;;  - Suite 330, Boston, MA 02111 USA.
;;;

;;; maths.asm: mathematical parser code
	
;;; Postfix conversion routines always take a single argument

ToAbc:	ld a,(iy+fmtFlags)
	and $ff^(1<<fmtBin)
	or (1<<fmtOct)|(1<<fmtHex)
	ld (iy+fmtOverride),a
	jr ToWhatever
ToDMS:	ld a,(iy+fmtFlags)
	or (1<<fmtBin)|(1<<fmtOct)|(1<<fmtHex)
	jr ToWhatever
ToDec:	ld a,(iy+fmtFlags)
	and $ff^((1<<fmtBin)|(1<<fmtOct)|(1<<fmtHex))
	ld (iy+fmtOverride),a
ToWhatever:
	call SetOP1Units
	jp nz,PopRetZ

	call ConvertToString

	jp PopRetNZ


;;; Addition/subtraction catch routines

Sub:	call StandardBinSet
	call BinarySetup
	jr nz,PopRetZ
	call CmpOP2OP4
	jp nz,ErrUnitMismatch
	ld a,(OP3)
	xor $80
	ld (OP3),a
	jr AddSub

Add:	call StringMathBinSet
	call BinarySetup
	jr nz,PopRetZ
AddSub:	call CmpOP2OP4
	jp nz,ErrUnitMismatch
	call OP3ToOP2
	call RealOP1OP2
	rst rFPAdd
BinaryDone:
	call OP4ToOP2
UnaryDone:
	call CplxOP1
	jr PopRetNZ

;;; These are used universally by random bits of code, since the zero flag is
;;; set or cleared by just about every routine in this app, to indicate status.

;;; Somewhat amusing story: when I added the Pop3Ret's, probably at about 3 AM
;;; during one of my coding sprees, I swapped the 'NZ' and 'Z' conditions...
;;; it took me around 2 weeks before I noticed anything, and three days before
;;; I tracked down the problem...

Pop3RetNZ:	pop hl
Pop2RetNZ:	pop hl
PopRetNZ:	pop hl
RetNZ:		or 1
		ret
		jr PopRetZ

Pop3RetZ:	pop hl
Pop2Ret:
Pop2RetZ:	pop hl
pzr:
PopRet:
PopRetZ:	pop hl
RetZ:		cp a
		ret

;;; Multiplication catch routine

Mul:	call StandardBinSet
	call BinarySetup
	jr nz,PopRetZ
	call AddUnitsOP2OP4
	call OP3ToOP2
	call RealOP1OP2
	B_CALL _FPMult
	jr BinaryDone


;;; Division catch routine

Div:	call StandardBinSet
	call BinarySetup
	jr nz,PopRetZ
	call SubUnitsOP2OP4
	call OP3ToOP2
	call RealOP1OP2
	B_CALL _FPDiv
	jr BinaryDone

UnitRecip:
	push hl
	jr RecipSt
Recip:	call SetOP1Units
	jr nz,PopRetZ
RecipSt:
	call OP2ToOP4
	call RealOP1
	B_CALL _FPRecip
	call OP4ToOP2
	call NegUnitsOP2
	jr UnaryDone

Square:	call SetOP1Units
	jr nz,PopRetZ
	call OP2ToOP4
	call AddUnitsOP2OP4
	call RealOP1
	B_CALL _FPSquare
	call OP4ToOP2
	jr UnaryDone

Cube:	call SetOP1Units
	jr nz,PopRetZ
	call OP2ToOP4
	call AddUnitsOP2OP4
	call AddUnitsOP2OP4
	call RealOP1
	B_CALL _Cube
	call OP4ToOP2
	jp UnaryDone

;;; Equalities:
;;; Should we demand that two numbers have the same units?
;;; Currently, all equalities except != will return 0 in this case.
;;; If you want to check from a BASIC program, you can use real( and imag(.

CompareUnitNumbers:
	;; Return values...
	;; NZ NC: Less
	;; NZ C:  Greater
	;; Z  NC: Equal
	;; Z  C:  Invalid comparison

	call BinarySetup
	jr nz,Pop2RetZ
	call CmpOP2OP4
	jr nz,CompareErr
	call RealOP1
	ld de,OP3
	ld hl,OP1
	ld a,(de)
	and $80
	cp (hl)
	jp nz,RetNZ
	inc hl
	inc de
;;; TI considers two numbers equal if their visible portions (first 10 digits)
;;; are equal. This sounds like a good cutoff, so I won't argue. Not checking
;;; the last few digits helps to prevent problems caused by rounding errors.
	ld b,6
	call CompStrsN
	ret
CompareErr:
	cp a
	scf
	ret

EQ:	call StringMathBinSet
	call CompareUnitNumbers
	;; EQ: Zero and not carry
	jr nz,EQ_Ret0
	jr c,EQ_Ret0
EQ_Ret1:
	B_CALL _OP1Set1
	jp PopRetNZ

NE:	call StringMathBinSet
	call CompareUnitNumbers
	;; NE: Nonzero (diff mantissa) or carry (diff units)
	jr nz,EQ_Ret1
	jr c,EQ_Ret1
EQ_Ret0:
	B_CALL _OP1Set0
	jp PopRetNZ

LT:	call CompareUnitNumbers
	;; LT: Nonzero and not carry
	jr z,EQ_Ret0
	jr c,EQ_Ret0
	jr EQ_Ret1

GT:	call CompareUnitNumbers
	;; GT: Nonzero and carry
	jr z,EQ_Ret0
GE_Nonzero:
	jr c,EQ_Ret1
	jr EQ_Ret0

LE:	call CompareUnitNumbers
	;; LE: Zero and not carry (equal) or nonzero and not carry (less)
GE_Zero:
	jr c,EQ_Ret0
	jr EQ_Ret1

GE:	call CompareUnitNumbers
	;; GE: Zero and not carry (equal) or nonzero and carry (greater)
	jr z,GE_Zero		; c=0,nc=1
	jr GE_Nonzero		; nc=0,c=1


;;; Exponentiation !!

;;; Allocation of registers:
;;;  OP7-OP9: used by reading code
;;;  OP7: stores the exponent
;;;  OP10: stores the current dimension of the base
;;;  OP11: temporarily used to store the scale of the base

CubeRt:	ld de,OneThird
	jr IRoot

Sqrt:	ld de,OneHalf
IRoot:	pop hl
	ld a,h
	or a
	jp nz,RetZ
	ld a,l
	dec a
	jp nz,RetZ
	push hl
	push de
	call SetOP1Units
	jp nz,Pop2RetZ
	call OP2ToOP10
	pop hl
	call Mov11OP2
	jr IRootPower


;;; Xroot: how annoying! Not only must we take the reciprocal, but the
;;; arguments are in the opposite order!

XRoot:	call StandardBinSet
	call SetOP1Units	; get the base
	jp nz,PopRetZ
	call OP2ToOP10		; save it while we handle the exponent
	call OP1ToOP11

	call PopOP1		; get the exponent
	ld a,(OP1)
	and $1f
	cp $0c
	jr z,XRootIsCplx	; complex exponent not allowed
	or a
	jp nz,PopRetZ
XRootIsReal:
	B_CALL _FPRecip
	rst rOP1ToOP2
	call OP11ToOP1
	jr XRootPower


XRootIsCplx:
	call SetOP1Units	; we already know it's complex; this is only
	ld hl,OP2+2		; to check whether the imaginary part is 0
	ld de,NoUnits+2
	ld b,7
	call CompStrsN
	jp nz,PopRetZ
	call RealOP1
	jr XRootIsReal

PowerIsCplx:
	call SetOP1Units	; same as above, only checking for imag=0
	ld hl,OP2+2
	ld de,NoUnits+2
	ld b,7
	call CompStrsN
	jp nz,PopRetZ
	call RealOP1
Power:	call StandardBinSet
	ld a,(OP1)		; exponent comes first in this case
	and $1f
	cp StrngObj
	jp z,PowerConv		; strings are conversions
	cp $0c
	jr z,PowerIsCplx	; same as for xroot, we don't want cpx exponent
	or a
	jp nz,PopRetZ

	call OP1ToOP11		; save exponent while we read the base

	call PopOP1
	call SetOP1Units
	jp nz,PowerFail1
	call OP2ToOP10
	call OP11ToOP2
XRootPower:
IRootPower:
	call OP2ToOP7		; save the exponent for later use
	call RealOP1
	B_CALL _YToX
	call OP1ToOP6		; and save the new scale

	ld b,7
	ld hl,OP10+2		; Now we process the dimension
PowerLoop:
	ld a,(hl)		; multiply each byte by the exponent
	cp $50
	jr z,Power_Skip
	push hl
	push bc
	jr nc,Power_Pos
	neg			; If it's negative, make it positive
	daa
	scf			; carry flag is set so we know it was negative
Power_Pos:
	push af

	sub $50			; remove the +50, make it nice to multiply
	daa

	ld hl,OP2
	ld (hl),0
	inc hl
	ld (hl),$81

	cp $10			; if our exponent is small, we must shift it
	jr nc,Power_NoShift	; over so that the first BCD digit is nonzero
	ld (hl),$80
	rlca
	rlca
	rlca
	rlca
Power_NoShift:
	inc hl			; Expand our single BCD byte into an entire
	ld (hl),a		; 9-digit number
	inc hl
	ld (hl),0
	ld de,OP2+4
	call Mov7	
	call OP7ToOP1
	B_CALL _FPMult		; and multiply it by the exponent
	pop af
	ld a,(OP1)
	rra			; The final sign of the byte will be that
	and $c0			; of the original byte xor that of our
	push af			; new result.
	xor a
	ld (OP1),a
	;; CkPosInt doesn't seem to do any rounding, which was the main reason
	;; I thought it might be worth using. So we are left to rounding it
	;; ourselves...
	ld d,9
	B_CALL _Round
	B_CALL _CkPosInt	; Is it valid?
	jp nz,ErrNonIntExpon
	ld a,(OP1+1)
	cp $82
	jp nc,ErrUnitOverflow
	cp $81
	ld a,(OP1+2)
	jr z,Power_Exp1
	rrca
	rrca
	rrca
	rrca
Power_Exp1:
	add a,$50
	daa
	jp c,ErrUnitOverflow
	ld b,a
	pop af
	ld a,b
	jp pe,Power_Pos_2	; Parity is even if we negated 0 or 2 times ^^
	xor a
	sub b
	daa
Power_Pos_2:
	pop bc
	pop hl
	ld (hl),a
Power_Skip:
	inc hl
	djnz PowerLoop
	call OP10ToOP2
	call OP6ToOP1
	jp UnaryDone

PowerFail1:
	call OP7ToOP1
	jp PopRetZ

PowerConv:
	;; We can also use a string (not actually numbers) as an exponent
	;; (which does nothing.) This is for "converting."
	call OP1ToOP11		; save exponent
	call PopOP1		; pop base
	call SetOP1Units	; convert / prefer base
	jp nz,PopRetZ
	ld de,OP10
	call Mov11FrOP1		; save base
	call OP11ToOP1		; restore exponent
	ld de,OP11
	call Mov11FrOP2		; save units of base
	call SetOP1Units	; convert / prefer exponent
	call OP1011ToOP12	; restore base
	jp UnaryDone


;;; Routines I haven't defined yet
;;; (through simple laziness)

Max	= pzr
Min	= pzr
Abs	= pzr
Sum	= pzr
Prod	= pzr


;;; Add one unit exponent

AddACUnits:
	sub $50
	daa
	rl b
	add a,c
	daa
	ld c,a

	;; This is ugly. We want to keep our numbers in the range 00-99, which
	;; corresponds to an exponent of -50 to +49.
	;;  If we carried zero or two times in the last two arithmetic
	;;  operations, the two carries cancel and we are still in the right
	;;  range.
	;;  If we carried once, we left the range on one end or the other and
	;;  didn't come back. So in this case we need to throw an error.
	ld a,0
	adc a,b
	rrca
	jp c,ErrUnitOverflow
	srl b
	ret

;;; Add units in OP2 to OP4

AddUnitsOP2OP4:
	ld hl,OP4+2
	ld de,OP2+2
AddUnits:
	ld b,7
AddUnitsLoop:
	ld a,(de)
	ld c,(hl)
	call AddACUnits
	ld (hl),c

	inc hl
	inc de
	djnz AddUnitsLoop
	ret

;;; Subtract units in OP4 from OP2; store in OP4

SubUnitsOP2OP4:
	ld hl,OP4+2
	ld de,OP2+2
SubUnits:
	ld b,7
SubUnitsLoop:
	ld a,(de)
	sub (hl)
	daa
	rl b
	add a,$50
	daa
	ld (hl),a

	;; See comment above for addition; the logic is the same.
	ld a,0
	adc a,b
	rrca
	jp c,ErrUnitOverflow
	srl b
	inc hl
	inc de
	djnz SubUnitsLoop
	ret

;;; Negate units

NegUnitsOP2:
	ld hl,OP2+2
NegUnits:
	ld b,7
NegUnitsLoop:
	ld a,$A0
	sub (hl)
	daa
	ld (hl),a
	inc hl
	djnz NegUnitsLoop
	ret


;;; Multiply units in OP4 by an integer

MulUnit:
	;; Unit * E
	ld hl,OP4+2
	ld c,7
MulUnitLoop:
	ld a,(hl)
	sub $50
	daa
	jr z,MulUnitSkip
	jr c,MulUnitNeg
	ld d,a
	call MulDE_BCD
	jp c,ErrUnitOverflow
	add a,$50
	daa
	jp c,ErrUnitOverflow
MulUnitNext:
	ld (hl),a
MulUnitSkip:
	inc hl
	dec c
	jr nz,MulUnitLoop
	ret

MulUnitNeg:
	ld a,$50
	sub (hl)
	daa
	ld d,a
	call MulDE_BCD
	jp c,ErrUnitOverflow
	ld d,a
	ld a,$50
	sub d
	daa
	jp c,ErrUnitOverflow
	jr MulUnitNext

MulDE_BCD:
	;; D is BCD, E is ordinary binary
	xor a
	ld b,8
MulDE_BCD_Loop:
	add a,a
	daa
	ret c
	sla e
	jr nc,MulDE_BCD_NC
	add a,d
	daa
	ret c
MulDE_BCD_NC:
	djnz MulDE_BCD_Loop
	ret

DivUnit:
	;; A (unit) / E (integer)
	;; Returns NZ if not nicely divisible
	ld d,a
	and $f0
	rrca
	ld b,a
	rrca
	rrca
	add a,b
	ld b,a
	ld a,d
	and $0f
	add a,e
	sub 50
	jr z,DivUnitPos
	neg
	call DivUnitPos
	ret nz
	neg
	add a,$a0
	cp a
	ret
DivUnitPos:
	ld d,a
	call DDivE
	or a
	ret nz
	ld e,10
	call DDivE
	ld b,a
	ld a,d
	rlca
	rlca
	rlca
	rlca
	add a,b
	add a,$50
	cp a
	ret	



DDivE:	xor a
	ld b,8
DDivEloop:
	sla     d
	rla
	cp      e
	jr      c,DDivEskip
	sub     e
	inc     d
DDivEskip:
	djnz DDivEloop
	ret


RealOP1OP2:
	ld a,(OP2)
	and $80
	ld (OP2),a
RealOP1:
	ld a,(OP1)
	and $80
	ld (OP1),a
	ret


CplxOP1:
	ld a,(OP1)
	or $0c
	ld (OP1),a
	ret


StringMathBinSet:
	;; In the cases of equality and addition, "A"+"B" actually means
	;; something in TIOS. Since addition of pure units has limited use,
	;; while breaking compatibility is a really bad idea, we want to let
	;; the OS handle this particular situation
	ld a,(OP1)
	and $1f
	jr z,StandardBinSet1
	cp StrngObj
	ret nz
	ld hl,(FPS)
	ld bc,-9
	add hl,bc
	ld a,(hl)
	and $1f
	cp StrngObj
	jp z,Pop2RetZ
	ret

StandardBinSet:
	;; No need for us to clutter up operations on two reals
	ld a,(OP1)
	and $1f
	ret nz
StandardBinSet1:
	ld hl,(FPS)
	ld bc,-9
	add hl,bc
	ld a,(hl)
	and $1f
	ret nz
	jp Pop2RetZ

BinarySetup:
	;; Input: OP1, FPST
	;; Output: if valid, OP3 and OP1, Z
	;; Uses all OP1-OP11.
	;; If not valid, arguments preserved and NZ set.
	call SetOP1Units
	ret nz
	call OP12ToOP1011
	call PopOP1
	call SetOP1Units
	jr nz,BinarySetupCancel
	ld hl,OP10
	ld de,OP3
	call Mov22
	cp a
	ret

BinarySetupCancel:
	;; We encountered something we didn't like
	;; so we must restore the first argument
	ld hl,OP11+2
	ld de,NoUnits+2
	ld b,7
	call CompStrsN
	jp nz,ErrDataType	; This would be bad.

	call OP10ToOP1
	call RealOP1
	or 1
	ret
