;;; -*- mode: TI-Asm; ti-asm-listing-file: "units.lst" -*-
;;;
;;;	Unit Operations Flash App for TI-83+
;;; 
;;;	Copyright 2003 Benjamin Moody
;;;
;;;  This program is free software; you can redistribute
;;;  and/or modify it under the terms of the GNU General
;;;  Public License as published by the Free Software
;;;  Foundation; either version 2 of the License, or (at
;;;  your option) any later version.
;;;
;;;  This program is distributed in the hope that it will be
;;;  useful, but WITHOUT ANY WARRANTY; without even the
;;;  implied warranty of MERCHANTABILITY or FITNESS FOR A
;;;  PARTICULAR PURPOSE. See the GNU General Public License
;;;  for more details.
;;;
;;;  You should have received a copy of the GNU General
;;;  Public License along with this program; if not, write
;;;  to the Free Software Foundation, Inc., 59 Temple Place
;;;  - Suite 330, Boston, MA 02111 USA.
;;;

;;; unittest.asm: an app that simply tests the validity of the data in
;;;    unitdata.asm by displaying all of the names. I used this quite a lot
;;;    to track down mistakes, usually an entry with the wrong name length.

;;; Include files omitted from listing
.nolist
#include "ti83plus.inc"
#include "extraequ.inc"
#include "hooks.inc"
.list

VMajor = 0
VMinor = 0

	.org $4000

	.db $80, $0F		; Field: Program length
	.db 0, 0, 0, 0
	.db $80, $12		; Field: Program type
	.db $01, $04
	.db $80, $21		; Field: App ID
	.db 1
	.db $80, $31		; Field: App Build
	.db 1
	.db $80, $48		; Field: App Name
	.db "UnitTest"
	.db $80, $81		; Field: App Pages
	.db 1
	.db $80, $90		; No default splash screen
	.db $03, $26, $09, $04, $04, $6F, $1B, $80	; Date stamp- 5/12/1999
	.db $02, $0D, $40, $A1, $6B, $99, $F6, $59, $BC, $67	; Dummy
	.db $F5, $85, $9C, $09, $6C, $0F, $B4, $03, $9B, $C9	; encrypted TI
	.db $03, $32, $2C, $E0, $03, $20, $E3, $2C, $F4, $2D	; date stamp
	.db $73, $B4, $27, $C4, $A0, $72, $54, $B9, $EA, $7C	; signature
	.db $3B, $AA, $16, $F6, $77, $83, $7A, $EE, $1A, $D4
	.db $42, $4C, $6B, $8B, $13, $1F, $BB, $93, $8B, $FC 
	.db $19, $1C, $3C, $EC, $4D, $E5, $75
	.db $80, $7F		; Field: Program Image length
	.db   0, 0, 0, 0
	.db $80,$A1		; Field: Hardware Compatibility Mode
	.db $FF

	.db 0,0,0,0
	.db 0,0,0,0
	.db 0,0,0,0
	.db 0

;;; Temporary storage locations (must be parserhook-safe!)

;;; Actually, many of these locations are used in both format and parse,
;;; but I'm too lazy to label them

unitPrefix	= appBackUpScreen	; Exponent of prefix
appVarPtr	= appBackUpScreen+1	; Pointer to start of prefs
lastPrefixPtr	= appBackUpScreen+3	; Pointer to prefix entry
helpPtr		= lastPrefixPtr		; Pointer to current help page
tempUnitPtr	= appBackUpScreen+5	; Pointer to unit entry
formatPtr	= appBackUpScreen+7	; Pointer to end of format string
parseNumDen	= appBackUpScreen+9	; Based on the previous separator
currentExponent = appBackUpScreen+10	; Exponent of a complete unit
tempStringMem	= appBackUpScreen+11
OP7		= appBackUpScreen+72	; Scratch space for FP
OP8		= appBackUpScreen+83
OP9		= appBackUpScreen+94
formatString	= appBackUpScreen+105	; Format string

;;; These locations must be rawkeyhook-safe
menuX		= saveSScreen
menuY		= saveSScreen+1
menuZ		= saveSScreen+2
menuSub		= saveSScreen+3
menuPtr		= saveSScreen+4
menuBot		= saveSScreen+6
tempLoadingOPs	= saveSScreen+7		; Used to save OPs during avar load


;;; Units are ALWAYS STORED IN SI BASE UNITS.
;;; There are seven of them, which is convenient for the 7 mantissa bytes...
;;; The order: meter, kilogram, second, ampere, kelvin, mole, candela

StartOApp:
	ld hl,0
	ld (curRow),hl
	B_CALL _ClrLCDFull
	set appAutoScroll,(iy+appFlags)
	ld hl,UnitStrings

Loop:	ld a,(hl)
	or a
	jr z,Quit

	ld a,$E0
	B_CALL _PutC

	ld b,(hl)		; Length of string
	inc hl
UnitMenuDrawItemTokenLoop:
	ld a,(hl)
	inc hl
	ld d,0
	ld e,a
	B_CALL _IsA2ByteTok
	jr nz,UnitMenuDrawItemTokenOne
	ld d,a
	ld e,(hl)
	inc hl
	dec b
UnitMenuDrawItemTokenOne:
	push bc
	 push hl
	  B_CALL _PutTokString
	  pop hl
	 pop bc
	djnz UnitMenuDrawItemTokenLoop
	
	ld a,$E0
	B_CALL _PutC

	B_CALL _NewLine

	push hl
	B_CALL _GetKey
	pop hl

	ld bc,15
	add hl,bc
	jr Loop
Quit:	B_CALL _JForceCmdNoChar


;;; Unit data
#include "unitdata.asm"

	.echo "Total length of app: "
	.echo $-$4000
	.echo "\n"

.end
END
